/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

//*	derived from Houndini sample: SOP_Surface

#ifndef __houdini_HoudiniSOP_h__
#define __houdini_HoudiniSOP_h__

namespace fe
{

/**************************************************************************//**
	@brief Houdini SOP node

	@ingroup houdini
*//***************************************************************************/
class HoudiniSOP:
	public SOP_Node,
	public MetaPlugin
{
	public:

	class OpOperator:
		public OP_Operator
	{
		public:
								OpOperator(const char *name,
										const char *english,
										OP_Constructor construct,
										PRM_Template *templates,
										unsigned min_sources,
										unsigned max_sources,
										CH_LocalVariable *variables,
										unsigned flags,
										const char **inputlabels,
										sp<Catalog> a_spCatalog);

								//* as OP_Operator
	virtual	bool				getOpHelpURL(UT_String& a_rUtString);

#if FE_HOUDINI_13_PLUS
    virtual bool				wantsInputEditor(void) const
								{	return false; }
#endif

		private:
			String				m_helpUrl;
	};

	class UndoCallback:
		public UT_Undo
	{
		public:
									UndoCallback(HoudiniSOP* a_pHoudiniSOP,
											String a_change,
											sp<Catalog> a_spCatalog,
											sp<Counted> a_spCounted);

	virtual	bool					isValid(void);
	virtual	void					undo(void)	{ undoOrRedo(FALSE); }
	virtual	void					redo(void)	{ undoOrRedo(TRUE); }

		private:

	virtual	void					undoOrRedo(BWORD a_redo);

			HoudiniSOP*				m_pHoudiniSOP;
			String					m_change;
			sp<Counted>				m_spCounted;
			sp<Catalog>				m_spCatalog;
	};

								HoudiniSOP(OP_Network *pNetwork,
										const char *name,OP_Operator *pOp);
virtual							~HoudiniSOP(void);

virtual	void					dirty(BWORD a_aggressive);
virtual	void					select(void);
virtual	BWORD					anticipate(String a_change,
										sp<Counted> a_spCounted);
virtual	BWORD					interrupted(void);

		BWORD					cookInputs(void);

static	OP_Node*				create(OP_Network *pNetwork,const char *name,
										OP_Operator *pOp);

static	int						discoverOperators(OP_OperatorTable *pTable);

								///	@internal
static	int						reloadCallback(void *pData,int index,float time,
										const PRM_Template* pTemplate);

								///	@internal
		void					brush(GU_Detail* pGdp,WindowEvent& a_rEvent,
										SpatialVector& a_rRayOrigin,
										SpatialVector& a_rRayDirection);

		void					clearErrors(void);
		void					relayErrors(void);
const	String					prompt(void);
const	String					idleEvents(void);
const	String					brushVisible(void);
const	String					brushCook(void);

	protected:
								//* as SOP_Node
virtual	OP_ERROR				bypassMe(OP_Context &context,int &copied_input);
virtual	void					opChanged(OP_EventType reason, void *data);

virtual void					getDescriptiveName(UT_String& a_rString) const;

virtual OP_ERROR				cookMySop(OP_Context& a_rContext);
virtual	OP_ERROR				cookMyGuide1(OP_Context& a_rContext);
virtual	OP_ERROR				save(std::ostream& a_rStreamOut,
										const OP_SaveFlags& a_rFlags,
										const char* a_pathPrefix);
virtual	bool					load(UT_IStream& a_rStreamIn,
										const char* a_extension,
										const char* a_path);

								//* as OP_Parameters
virtual	unsigned int			disableParms(void);

	private:
		void					createOperator(String a_operatorName,
										Real a_time);
		void					updateCatalog(Real time,BWORD a_doInput,
										BWORD a_force=FALSE);
		void					resignalBrush(void);

static	int						buttonCallback(void* a_pNode,int a_index,
										fpreal32 a_time,
										const PRM_Template* a_pPrmTemplate);

		HoudiniContext					m_houdiniContext;
		String							m_brushName;
		String							m_loadedState;
		double							m_frame;
		BWORD							m_cooking;
		BWORD							m_inputsLocked;
		BWORD							m_guideEmpty;
		BWORD							m_temporaryTemplate;

		OP_Operator*					m_pOpOperator;
		OP_Context*						m_pOpContext;
		UT_AutoInterrupt*				m_pInterrupt;
};

} // namespace

#endif /* __houdini_HoudiniSOP_h__ */
