/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __oplab_SurfaceViewerOp_h__
#define __oplab_SurfaceViewerOp_h__

namespace fe
{

/**************************************************************************//**
	@brief Operator to look at surfaces in an independent window

	@ingroup oplab
*//***************************************************************************/
class FE_DL_EXPORT SurfaceViewerOp:
	public OperatorSurfaceCommon,
	public Initialize<SurfaceViewerOp>
{
	public:

							SurfaceViewerOp(void);
virtual						~SurfaceViewerOp(void);

		void				initialize(void);

							//* As HandlerI
virtual	void				handle(Record& a_rSignal);

		sp<QuickViewerI>	quickViewer(void)		{ return m_spQuickViewerI; }
		sp<DrawableI>		drawable1(void)			{ return m_spDrawable1; }
		sp<DrawableI>		drawable2(void)			{ return m_spDrawable2; }
		sp<DrawableI>		drawable3(void)			{ return m_spDrawable3; }
		sp<DrawableI>		drawable4(void)			{ return m_spDrawable4; }

		DrawMode::DrawStyle	drawStyle(void) const	{ return m_drawStyle; }
		Real				lineWidth(void) const	{ return m_lineWidth; }

		Real				updateSlack(void);
		void				setSlack(Real a_slack,Real a_slackInc,
								Real a_slackMin)
							{	m_slack=a_slack;
								m_slackInc=a_slackInc;
								m_slackMin=a_slackMin; }

		String				findPartitionAttribute(sp<SurfaceAccessibleI>
									a_spSurfaceAccessible);

	private:

		void				post(Real a_frame);
		void				react(WindowEvent& a_rEvent);

	class DrawHandler: virtual public HandlerI
	{
		public:
						DrawHandler(void);

			void		bind(sp<DrawI> a_spDrawI)
						{	m_spDrawI=a_spDrawI; }

			void		bindDrawable1(sp<DrawableI> a_spDrawable1)
						{ m_spDrawable1=a_spDrawable1; }
			void		bindDrawable2(sp<DrawableI> a_spDrawable2)
						{ m_spDrawable2=a_spDrawable2; }
			void		bindDrawable3(sp<DrawableI> a_spDrawable3)
						{ m_spDrawable3=a_spDrawable3; }
			void		bindDrawable4(sp<DrawableI> a_spDrawable4)
						{ m_spDrawable4=a_spDrawable4; }

			void		setFrame(Real a_frame);
			void		setMicroseconds(I32 a_us)	{ m_us=a_us; }

			sp<SurfaceI> impactSurface(void)
						{	return m_spImpactSurface; }
			sp<SurfaceI> surface(I32 a_index);

			I32			isolatedIndex(void)
						{	return m_isolateSurfaceIndex; }

			BWORD		hideAll(I32 a_surfaceIndex)
						{	return m_hideAll[a_surfaceIndex]; }
			BWORD		ghostAll(I32 a_surfaceIndex)
						{	return m_ghostAll[a_surfaceIndex]; }

			void		setSurfaceViewerOp(
								sp<SurfaceViewerOp> a_spSurfaceViewerOp)
						{	m_hpSurfaceViewerOp=a_spSurfaceViewerOp; }
			void		setImpactAccessible(
							sp<SurfaceAccessibleI> a_spSurfaceAccessible)
						{	m_spImpactAccessible=a_spSurfaceAccessible; }
			void		setImpactSurface(sp<SurfaceI> a_spSurface)
						{	m_spImpactSurface=a_spSurface; }
			void		setSurfaceImpact(
								sp<SurfaceI::ImpactI> a_spSurfaceImpact)
						{	m_spSurfaceImpact=a_spSurfaceImpact; }
			void		setImpactSurfaceName(String a_surfaceName)
						{	m_impactSurfaceName=a_surfaceName; }
			void		setImpactSurfaceIndex(I32 a_surfaceIndex)
						{	m_impactSurfaceIndex=a_surfaceIndex; }

			void		unpick(void);
			void		hideSelected(void);
			void		ghostSelected(void);
			void		isolateToggle(void);

			BWORD		getPickRegion(SpatialVector& a_rCenter,
							Real& a_rRadius) const ;

	virtual void		handle(Record& a_render);

			void		applyHidden(I32 a_surfaceIndex);
			void		applyGhosted(I32 a_surfaceIndex);

		private:

			AsViewer				m_asViewer;
			sp<DrawI>				m_spDrawI;
			sp<DrawableI>			m_spDrawable1;
			sp<DrawableI>			m_spDrawable2;
			sp<DrawableI>			m_spDrawable3;
			sp<DrawableI>			m_spDrawable4;
			Real					m_frame;
			I32						m_us;

			BWORD					m_pickRegion;
			SpatialVector			m_pickCenter;
			Real					m_pickRadius;
			String					m_pickName;
			BWORD					m_hideAll[4];
			BWORD					m_ghostAll[4];
			Array<String>			m_hideList[4];
			Array<String>			m_ghostList[4];
			sp<PartitionI>			m_spGhostPartition[4];
			I32						m_isolateSurfaceIndex;
			String					m_isolatePart;

			BWORD					m_keepEdges;
			std::map<U64,Edge>		m_edgeMap;

			hp<SurfaceViewerOp>		m_hpSurfaceViewerOp;
			sp<SurfaceAccessibleI>	m_spImpactAccessible;
			sp<SurfaceI>			m_spImpactSurface;
			sp<SurfaceI::ImpactI>	m_spSurfaceImpact;
			String					m_impactSurfaceName;
			I32						m_impactSurfaceIndex;
			String					m_partAttr;

			sp<DrawMode>			m_spDrawGhost;
			sp<DrawMode>			m_spDrawPoints;
			sp<DrawMode>			m_spDrawWire;
			sp<DrawMode>			m_spDrawImpact;
			sp<DrawMode>			m_spDrawOutline;
	};

	class EventHandler: virtual public HandlerI
	{
		public:
					EventHandler(void)										{}

	virtual void	handle(Record& a_render);

			void	setSurfaceViewerOp(sp<SurfaceViewerOp> a_spSurfaceViewerOp)
					{	m_hpSurfaceViewerOp=a_spSurfaceViewerOp; }

		private:
			WindowEvent			m_event;
			hp<SurfaceViewerOp>	m_hpSurfaceViewerOp;
	};

	class Worker: public Thread::Functor
	{
		public:
							Worker(sp< JobQueue<Real> > a_spJobQueue,
								U32 a_id,String a_stage):
								m_hpJobQueue(a_spJobQueue),
								m_id(a_id),
								m_us(0)										{}

	virtual					~Worker(void)									{}

	virtual	void			operate(void);

			void			setObject(sp<Component> a_spObject)
							{	m_hpSurfaceViewerOp=a_spObject; }

		private:
			hp< JobQueue<Real> >	m_hpJobQueue;
			U32						m_id;
			U32						m_us;

			hp<SurfaceViewerOp>		m_hpSurfaceViewerOp;
	};

		U32							m_jobs;
		sp< JobQueue<Real> >		m_spJobQueue;
		sp< Gang<Worker,Real> >		m_spGang;
		sp<QuickViewerI>			m_spQuickViewerI;
		sp<EventHandler>			m_spEventHandler;
		sp<DrawHandler>				m_spDrawHandler;
		sp<RecorderI>				m_spVideoRecorder;
		sp<DrawableI>				m_spDrawable1;
		sp<DrawableI>				m_spDrawable2;
		sp<DrawableI>				m_spDrawable3;
		sp<DrawableI>				m_spDrawable4;

		DrawMode::DrawStyle			m_drawStyle;
		Real						m_lineWidth;
		Real						m_lastFrame;
		Real						m_slack;
		Real						m_slackInc;
		Real						m_slackMin;

	public:

		sp<DrawHandler>		drawHandler(void)	{ return m_spDrawHandler; }
};

} /* namespace */

#endif /* __oplab_SurfaceViewerOp_h__ */
