/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "surface/surface.pmh"

namespace fe
{

SurfaceAccessibleCatalog::SurfaceAccessibleCatalog(void):
	m_persistant(FALSE)
{
}

SurfaceAccessibleCatalog::~SurfaceAccessibleCatalog(void)
{
	//* TODO consider reference counting
	if(!m_persistant)
	{
		reset();
	}
}

sp<Catalog> SurfaceAccessibleCatalog::catalog(void)
{
	if(!m_spCatalog.isValid())
	{
		//* auto-bind a private catalog
		m_spCatalog=registry()->master()->createCatalog("Surface Catalog");
	}

	return m_spCatalog;
}

void SurfaceAccessibleCatalog::reset(void)
{
	sp<Catalog> spCatalog=catalog();

	if(spCatalog.isValid())
	{
		spCatalog->catalogRemove(m_key);
	}
}

void SurfaceAccessibleCatalog::attributeSpecs(
	Array<SurfaceAccessibleI::Spec>& a_rSpecs,
	SurfaceAccessibleI::Element a_element) const
{
//	feLog("SurfaceAccessibleCatalog::attributeSpecs element %d\n",a_element);

	a_rSpecs.clear();

	if(!m_spCatalog.isValid())
	{
		return;
	}

	const String elementName=elementLayout(a_element);

	Array<String> properties;
	m_spCatalog->catalogProperties(m_key,properties);

	const U32 propertyCount=properties.size();
	for(U32 propertyIndex=0;propertyIndex<propertyCount;propertyIndex++)
	{
		const String attrName=properties[propertyIndex].replace("\\[0\\]$","");

		if(attrName=="surf:points")
		{
			continue;
		}

		if(!strncmp(attrName.c_str(),"type:",5) ||
				!strncmp(attrName.c_str(),":",1) )
		{
			continue;
		}

		const String rateName=m_spCatalog->catalogOrDefault<String>(
				m_key,"rate:"+attrName,"");
		if(rateName!=elementName)
		{
			continue;
		}

		const String typeName=m_spCatalog->catalogOrDefault<String>(
				m_key,"type:"+attrName,"");

//		feLog("SurfaceAccessibleCatalog::attributeSpecs %d/%d"
//				" rate \"%s\" type \"%s\" attr \"%s\"\n",
//				propertyIndex,propertyCount,
//				rateName.c_str(),typeName.c_str(),attrName.c_str());

		SurfaceAccessibleI::Spec spec;
		spec.set(commonName(attrName),typeName);

		a_rSpecs.push_back(spec);
	}
}

sp<SurfaceAccessorI> SurfaceAccessibleCatalog::accessor(
	SurfaceAccessibleI::Element a_element,
	String a_name,
	SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorCatalog> spAccessor=accessor();
	if(spAccessor.isValid() && (spAccessor->bind(a_element,a_name) ||
			a_create==SurfaceAccessibleI::e_createMissing))
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		return spAccessor;
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleCatalog::accessor(
	SurfaceAccessibleI::Element a_element,
	SurfaceAccessibleI::Attribute a_attribute,
	SurfaceAccessibleI::Creation a_create)
{
	sp<SurfaceAccessorCatalog> spAccessor=accessor();
	if(spAccessor.isValid() && (spAccessor->bind(a_element,a_attribute) ||
			a_create==SurfaceAccessibleI::e_createMissing))
	{
		spAccessor->setSurfaceAccessible(sp<SurfaceAccessibleI>(this));
		return spAccessor;
	}
	return sp<SurfaceAccessorI>(NULL);
}

sp<SurfaceAccessorI> SurfaceAccessibleCatalog::accessor(void)
{
	sp<Catalog> spCatalog=catalog();

	if(!spCatalog.isValid())
	{
		return sp<SurfaceAccessorI>(NULL);
	}
	sp<SurfaceAccessorCatalog> spAccessorCatalog(new SurfaceAccessorCatalog());
	if(spAccessorCatalog.isValid())
	{
		spAccessorCatalog->setCatalog(spCatalog);

		//* NOTE intentionally duplicating String buffer to be thread safe
		spAccessorCatalog->setKey(m_key.c_str());
	}

	return spAccessorCatalog;
}

BWORD SurfaceAccessibleCatalog::discard(
	SurfaceAccessibleI::Element a_element,String a_name)
{
	if(!m_spCatalog.isValid())
	{
		return FALSE;
	}

	m_spCatalog->catalogRemove(m_key,a_name);
	return TRUE;
}

} // namespace
