/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceAccessibleRecord_h__
#define __surface_SurfaceAccessibleRecord_h__

namespace fe
{

/**************************************************************************//**
    @brief Record Surface Binding

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleRecord:
	public ObjectSafeShared<SurfaceAccessibleRecord>,
	virtual public SurfaceAccessibleBase
{
	public:
								SurfaceAccessibleRecord(void);
virtual							~SurfaceAccessibleRecord(void);

								//* As Protectable
virtual	Protectable*			clone(Protectable* pInstance=NULL);

								//* as SurfaceAccessibleI
virtual	void					bind(Instance a_instance)
								{
									Record record=a_instance.cast<Record>();
									if(record.isValid())
									{
										sp<Scope> spScope=
												record.layout()->scope();
										bind(spScope);
									}
									bind(record);
								}
virtual	BWORD					isBound(void)
								{	return m_weakRecord.isValid(); }

virtual	void					bind(sp<Scope> a_spScope)
								{	m_spScope=a_spScope;
									reset(); }
virtual	BWORD					load(String a_filename)
								{	return load(a_filename,sp<Catalog>(NULL)); }
virtual	BWORD					load(String a_filename,
										sp<Catalog> a_spSettings);
virtual	BWORD					save(String a_filename)
								{	return save(a_filename,sp<Catalog>(NULL)); }
virtual	BWORD					save(String a_filename,
										sp<Catalog> a_spSettings);

virtual	void					bind(sp<SurfaceAccessibleI::ThreadingState>
										a_spThreadingState)
								{	m_spThreadingState=a_spThreadingState; }

virtual	void					groupNames(Array<String>& a_rNameArray,
										SurfaceAccessibleI::Element
										a_element) const;

virtual	void					attributeSpecs(
										Array<SurfaceAccessibleI::Spec>&
										a_rSpecs,
										SurfaceAccessibleI::Element
										a_element) const;

virtual void					require(SurfaceAccessibleI::Element a_element,
										String a_name,String a_typeName);

virtual	I32						count(SurfaceAccessibleI::Element
										a_element) const;

virtual sp<SurfaceAccessorI>	accessor(Element a_element,String a_name,
										Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(Element a_element,
										Attribute a_attribute,
										Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(Element a_element,String a_name)
								{	return accessor(a_element,a_name,
											e_createMissing); }

virtual sp<SurfaceAccessorI>	accessor(Element a_element,
										Attribute a_attribute)
								{	return accessor(a_element,a_attribute,
											e_createMissing); }

virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										String a_name)
								{	return FALSE; }
virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute);

virtual	sp<SurfaceI>			surface(void)
								{	return surface(String()); }
virtual	sp<SurfaceI>			surface(String a_group);
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex)
								{	return subSurface(a_subIndex,String()); }
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group);

virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible)
								{	reset();
									append(a_spSurfaceAccessible,NULL); }

virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										sp<FilterI> a_spFilter)
								{	reset();
									append(a_spSurfaceAccessible,
											NULL,a_spFilter); }

virtual	void					instance(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										const Array<SpatialTransform>&
										a_rTransformArray)					{}

								//* default implementation is non-atomic
virtual	sp<SpannedRange>		atomize(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount);

virtual	sp<SpannedRange>		atomizeConnectivity(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount,
										BWORD a_checkPages);

								//* as SurfaceAccessibleRecord
		void					bind(Record& a_rRecord);
		Record					record(void)
								{	return m_weakRecord; }

		void					bindLike(sp<SurfaceAccessibleRecord>& a_rspOther)
								{
									if(a_rspOther.isValid())
									{
										m_spThreadingState=
												a_rspOther->m_spThreadingState;
										m_paging=a_rspOther->m_paging;
										m_spScope=a_rspOther->m_spScope;
										m_spRecordGroup=
												a_rspOther->m_spRecordGroup;
										m_weakRecord=a_rspOther->m_weakRecord;
									}
								}

	protected:
virtual	void					reset(void);

		sp<Scope>				m_spScope;
		sp<RecordGroup>			m_spRecordGroup;
		WeakRecord				m_weakRecord;
};

} // namespace

#endif /* __surface_SurfaceAccessibleRecord_h__ */
