import sys
import os
import re
forge = sys.modules["forge"]

def setup(module):

	manifest = 'ext/xgen/manifest.cc'
	with open(manifest, 'w') as outfile:
		outfile.write("")

	for maya_source in reversed(sorted(forge.maya_sources)):
		maya_version = maya_source[0]

		# NOTE Maya 2016 has broken XpdFile.h (no such xpd/Xpd.h)
#		if re.match("2016", maya_version):
#			continue

		maya_root = maya_source[1] + "/.."

		maya_xgen = maya_root + "/plug-ins/xgen"
		if not os.path.exists(maya_xgen):
			continue

		sys.stdout.write(" " + maya_version)

		xgen_major = maya_version.split(".")[0].split("sp")[0].split("pr")[0]

		srcList = [	"SurfaceAccessibleXGen",
					"xgen.pmh",
					"xgenDL" ]

		variant = maya_version
		if variant != "":
			variantPath = module.modPath + '/' + variant
			if os.path.lexists(variantPath) == 0:
				os.symlink('.', variantPath)

			srcListVariant = []
			for src in srcList:
				srcListVariant += [ variant + '/' + src ]
			srcList = srcListVariant

		dllname = "fexXGenDL" + variant

		with open(manifest, 'a') as outfile:
			suffix = ""
			if variant != "":
				suffix = "." + variant
			outfile.write('\tspManifest->catalog<String>(\n'+
					'\t\t\t"SurfaceAccessibleI.SurfaceAccessibleXGen.fe.xgen'+
					suffix+'")=\n'+
					'\t\t\t"fexXGenDL'+variant+'";\n');

		dll = module.DLL(dllname, srcList)

		for src in srcList:
			relativePath = forge.RelativePath(forge.rootPath, module.modPath)
			modObjPath = os.path.join(forge.objPath, relativePath)
			objName = modObjPath + '/' + src
			if re.compile(r'(.*)\.pmh$').match(src):
				objName += '.gch'
			else:
				objName += forge.objsuf
			srcTarget = forge.FindTarget(objName)

			srcTarget.cppmap = { "xgen" : " -D XGEN_MAJOR=" + xgen_major }

			srcTarget.includemap = {}
			srcTarget.includemap['maya'] = maya_root + "/include"
			srcTarget.includemap['xgen'] = maya_xgen + "/include"
			srcTarget.includemap['xgen2'] = maya_xgen + "/include/xgen"
			srcTarget.includemap['xgen3'] = maya_xgen + "/include/XGen"

			# need std::shared::ptr
			srcTarget.cppmap["std"] = "-std=c++0x"

		maya_lib = maya_root + "/lib"
		xgen_lib = maya_xgen + "/lib"

		dll.linkmap = {}

		dll.linkmap['xgen'] = " -Xlinker -rpath '" + xgen_lib + "/'"
		dll.linkmap['xgen'] += " -L" + xgen_lib + " -lAdskXGen"
#		dll.linkmap['xgen'] += " -L" + xgen_lib + " -lAdskXpd"
#		dll.linkmap['xgen'] += " -L" + xgen_lib + " -lAdskSeExpr"

		# Ptex is not directly used, but seems to get rpath to work
		dll.linkmap['xgen'] += " -Xlinker -rpath '" + maya_lib + "/'"
		dll.linkmap['xgen'] += " -L" + maya_lib + " -lPtex"

		deplibs = forge.corelibs+ [
					"fexSurfaceDLLib" ]

		forge.deps( [ dllname + "Lib" ], deplibs )

	forge.tests += [
		("inspect.exe",		"fexXGenDL",	None,		None) ]

	module.Module('test')
